from . import bi_avg
from . import cumulative
from . import exp_smooth
from . import pid
from . import ring_buff
from . import smooth

__all__ = [
    "bi_avg",
    "cumulative",
    "exp_smooth",
    "pid",
    "ring_buff",
    "smooth",
]


import typing as t


class Stat:

    def __init__(self) -> None:
        self.value = 0.0
        self.count = 0

    def __str__(self) -> str:
        return "Average"

    def add(self, value: float):
        self.value += value
        self.count += 1

    def get(self) -> t.Tuple[float, float, float]:
        if self.count == 0:
            return 0
        return self.value / self.count



import typing as t


class Stat:

    def __init__(self, alpha: float) -> None:
        self.hit_avg = 0.0
        self.average = 0.0
        self.duration_miss = 0.0

        self.alpha = 1
        self.steps = 1
        if alpha > 1:
            self.steps = alpha
            self.is_avg = True
        elif alpha <= 1:
            self.alpha = alpha
            self.is_avg = False

    def __str__(self) -> str:
        return f"ExpSmooth <α:({self.alpha},s:{self.steps})>"

    def simgle_average(self, value: float):
        return (self.average * (self.steps - 1) + value) / self.steps

    def exponential_average(self, value: float):
        return (self.alpha * value + (1 - self.alpha) * self.average)

    def add(self, value: float):
        if self.is_avg:
            self.average = self.simgle_average(value)
        else:
            self.average = self.exponential_average(value)

    def get(self) -> t.Tuple[float, float, float]:
        return self.average




class Stat:
    def __init__(self, alpha: float = 0.5, beta: float = 0.3, gamma: float = 0.2, integral_limit: float = 200):
        """
        Initialize Stats with PID-like smoothing parameters.
        :param alpha: Proportional coefficient.
        :param beta: Integral coefficient.
        :param gamma: Derivative coefficient.
        """

        # alpha + beta should ber equal 1 for better percision
        if not (0 < alpha <= 1 and 0 <= beta <= 1 and 0 <= gamma <= 1):
            raise ValueError("alpha, beta, and gamma must be in the range [0, 1] and alpha > 0")

        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma
        self.integral_limit = integral_limit

        self.value_integral = 0.0
        self.value_prev = None
        self.value_avg = 0.0

    def __str__(self) -> str:
        return f'PID <α:{self.alpha},β:{self.beta},γ:{self.gamma},δ:{self.integral_limit}>'

    def add(self, value: float):
        """
        Update hit statistics with a new duration.
        :param start: Start time in seconds.
        :param end: End time in seconds.
        """

        # Update integral component
        # self.value_integral += duration
        self.value_integral = (
            self.value_integral * (self.integral_limit - 1) + value
        ) / self.integral_limit

        # Update derivative component
        derivative = 0.0
        if self.value_prev is not None:
            derivative = value - self.value_prev

        # Calculate new average with PID components
        self.value_avg = (
            self.alpha * value +
            self.beta * self.value_integral +
            self.gamma * derivative
        )

        # Store current value for next derivative calculation
        self.value_prev = value

    def get(self) -> float:
        """
        Get the current average value.
        :return: Average value.
        """
        return self.value_avg



class RingBuff:
    def __init__(self, size=100) -> None:
        self.buff = [0] * size
        self.size = size
        self.sum = 0
        self.count = 0
        self.index = 0

    def Add(self, value: int):
        if self.count >= self.size:
            self.sum -= self.buff[self.index]
        else:
            self.count += 1

        self.buff[self.index] = value
        self.sum += value

        self.index = (self.index + 1) % self.size

    def Average(self) -> float:
        if self.count == 0:
            return 0.0
        return self.sum / float(self.count)


class Stat:
    def __init__(self, size: int) -> None:
        self.size = size
        self.value = RingBuff(size)

    def __str__(self) -> str:
        return f"RingBuff <size:{self.size}>"

    def add(self, value: float):
        self.value.Add(value)

    def get(self) -> float:
        return self.value.Average()



from threading import Lock


class Stat:

    def __init__(self, alpha: float = 0.3, beta: float = 0.2):
        """
        Initialize Stats with smoothing parameters.

        :param alpha: Smoothing factor for average values.
        :param beta: Smoothing factor for trend values.
        """
        if not (0 < alpha <= 1 and 0 < beta <= 1):
            raise ValueError("alpha and beta must be in the range (0, 1]")

        self.alpha = alpha
        self.beta = beta

        self.avg = 0.0
        self.trend = 0.0

        self.lock = Lock()

    def __str__(self) -> str:
        return f"BiAvg <α:{self.alpha}, β:{self.beta}>"

    def add(self, value: float):
        """
        Update count statistics with a new duration.
        """
        with self.lock:
            prev_avg = self.avg
            self.avg = (self.alpha * value + (1 - self.alpha) *
                        (self.avg + self.trend))
            self.trend = (self.beta * (self.avg - prev_avg) +
                          (1 - self.beta) * self.trend)

    def get(self) -> float:
        """
        Get the current average count.
        :return: Average count.
        """
        return self.avg



import math
from time import monotonic_ns


class Stat:

    def __init__(self, tau=60):
        self.hit_rate = 0.0
        self.value_avg = 0.0
        self.miss_avg_duration = 0.0
        self.last_update = 0
        self.tau = tau
        self.total_hits = 0
        self.total_misses = 0

    def __str__(self) -> str:
        return f'Smooth <τ:{self.tau}>'

    def _update_smooth(self, old_value, current_value, elapsed_time):
        exp_factor = math.exp(-elapsed_time / self.tau)
        return old_value * exp_factor + current_value * (1 - exp_factor)

    def add(self, value: float):
        now = monotonic_ns()
        time_since_update = now - self.last_update

        # Update smoothed averages
        self.value_avg = self._update_smooth(self.value_avg, value, time_since_update)

        # Update last update timestamp
        self.last_update = now

    def get(self):
        return self.value_avg
